"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Gateway = void 0;
const cosmwasm_stargate_1 = require("@cosmjs/cosmwasm-stargate");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const bs58_1 = __importDefault(require("bs58"));
const address_1 = require("./address");
const platform_1 = require("./platform");
const constants_1 = require("./constants");
var Gateway;
(function (Gateway) {
    Gateway.name = "Wormchain";
    function gatewayAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.gateway;
    }
    Gateway.gatewayAddress = gatewayAddress;
    function tokenBridgeAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.tokenBridge;
    }
    Gateway.tokenBridgeAddress = tokenBridgeAddress;
    function coreAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.coreBridge;
    }
    Gateway.coreAddress = coreAddress;
    // Returns RPC client for Wormchain
    async function getRpc() {
        const rpcAddress = platform_1.CosmwasmPlatform.conf[Gateway.name].rpc;
        return await cosmwasm_stargate_1.CosmWasmClient.connect(rpcAddress);
    }
    Gateway.getRpc = getRpc;
    async function getTokenBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return platform_1.CosmwasmPlatform.getTokenBridge(rpc);
    }
    Gateway.getTokenBridge = getTokenBridge;
    async function getIbcBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return platform_1.CosmwasmPlatform.getIbcBridge(rpc);
    }
    Gateway.getIbcBridge = getIbcBridge;
    // Get the wrapped version of an asset created on wormchain
    // for a given chain
    async function getWrappedAsset(token) {
        const tb = await getTokenBridge();
        const wrappedAsset = await tb.getWrappedAsset(token);
        // Encode the original address to base58 and add it
        // to the factory address for cw20 style factory token address
        const encodedAddress = bs58_1.default.encode(wrappedAsset.toUniversalAddress().toUint8Array());
        const factoryAddress = `factory/${gatewayAddress()}/${encodedAddress}`;
        return new address_1.CosmwasmAddress(factoryAddress);
    }
    Gateway.getWrappedAsset = getWrappedAsset;
    // Gets the the source channel for outgoing transfers from wormchain
    function getGatewaySourceChannel(chain) {
        const channels = platform_1.CosmwasmPlatform.getIbcChannels(chain);
        if (!channels)
            throw new Error("No channels configured for chain " + chain);
        if (!(Gateway.name in channels))
            throw new Error("No channel configured for chain " + chain);
        return channels[Gateway.name];
    }
    Gateway.getGatewaySourceChannel = getGatewaySourceChannel;
    // derive the ics20 token denom from the
    // wrapped denom and destination channel
    function deriveIbcDenom(chain, denom) {
        // Otherwise compute the ibc address from the channel and denom
        const channel = getGatewaySourceChannel(chain);
        const hashData = connect_sdk_1.encoding.toUint8Array(`${constants_1.IBC_TRANSFER_PORT}/${channel}/${denom}`);
        const hash = connect_sdk_1.encoding.hex.encode((0, connect_sdk_1.sha256)(hashData));
        return new address_1.CosmwasmAddress(`ibc/${hash.toUpperCase()}`);
    }
    Gateway.deriveIbcDenom = deriveIbcDenom;
    function makeTransferMsg(chain, recipient, fee = 0n, payload, nonce) {
        // Address of recipient is b64 encoded Cosmos bech32 address
        const address = connect_sdk_1.encoding.b64.encode(recipient.toUint8Array());
        const common = {
            chain: (0, connect_sdk_1.toChainId)(chain),
            recipient: address,
            fee: fee.toString(),
            nonce: nonce ?? Math.round(Math.random() * 100000),
        };
        const msg = payload
            ? {
                gateway_transfer_with_payload: { ...common, payload: payload },
            }
            : { gateway_transfer: { ...common } };
        return msg;
    }
    Gateway.makeTransferMsg = makeTransferMsg;
    // TODO: make consts
    function parseWormholeMessage(tx) {
        const events = tx.events.filter((ev) => ev.type === "wasm" &&
            ev.attributes[0].key === "_contract_address" &&
            ev.attributes[0].value === coreAddress());
        if (events.length === 0)
            throw new Error("No wormhole message found in tx");
        if (events.length > 1)
            console.error(`Expected single message, found ${events.length}`);
        const [wasm] = events;
        const sequence = wasm.attributes.find((e) => {
            return e.key === "message.sequence";
        }).value;
        const emitter = wasm.attributes.find((e) => {
            return e.key === "message.sender";
        }).value;
        return {
            chain: Gateway.name,
            emitter: new connect_sdk_1.UniversalAddress(emitter),
            sequence: BigInt(sequence),
        };
    }
    Gateway.parseWormholeMessage = parseWormholeMessage;
})(Gateway || (exports.Gateway = Gateway = {}));
//# sourceMappingURL=gateway.js.map